/*****************************************************************************
    cache.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2005 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include "globals.h"
#ifdef FILE_CACHE
#include "util.h"
#include "netio.h"
#include "fileio.h"
#include "logfun.h"
#include "cache.h"
#include "realpath.h"

static char *next = NULL;
static int  avail = 0;

/*
	check the current cache limits
*/

int cacheLimits(ulong size, char *name)
{
    if (!maxFilecache)
        return 1;

    if ((size > maxCacheobj) || ((size + currCachesize) > maxFilecache)) {
        LOG_MSG3 (LOG_NOTICE, _("File '%s' too large for cache"), 
                  baseName(name, 0));
        return 1;
    }
    return 0;
}

/*
	scan a direcctory for a matching plain file
	and add it to the cache list
*/

int scanDir(char *dir, regex_t *compiled, int flag)
{
    DIR *dp;
    struct dirent *de;
    struct stat sb;
    char *new;

    if ((dp = opendir(dir)) == NULL) {
        return 1;
    }

    LOG_MSG6 (LOG_DEBUG, "Scanning directory '%s'", dir);

    while ((de = readdir(dp)) != NULL) {
        if (!strcmp(de->d_name,"."))
            continue;
        if (!strcmp(de->d_name,".."))
            continue;

        new = Malloc(strlen(dir)+strlen(de->d_name)+2);
        sprintf(new,"%s/%s",dir,de->d_name);

        if (stat(new, &sb))
            continue;

        if (S_ISREG(sb.st_mode)) {
            if (regexec(compiled, new, 0, NULL, 0))
                continue;
            if (cacheLimits((ulong)sb.st_size, new))
                continue;
            if (addCache(cacheListH, (flag & CAC_MODE), new, &sb, 
                         (flag & CAC_PROP), NULL)) {
                errNoDynmem;
                return -1;
            }
            LOG_MSG5 (LOG_DEBUG, 
                      "File '%s' (flags: 0x%x) marked for cache",
                      de->d_name, flag);
        }
        else if ((flag && CAC_RECURSIVE) && (S_ISDIR(sb.st_mode))) {
            scanDir(new, compiled, flag);
            LOG_MSG6 (LOG_DEBUG, "Back in directory '%s'", dir);
        }

        free(new);
    }
    closedir(dp);
    return 0;
}

/*
	list of eligible cache objects and scan the directories for
	preloadable files. this function does the scan and cache list
	update only. the real thing is done by preloadCache().
*/

int addCacheObjs(void)
{
    dlr_t *dr;
    char *dir;
    int rtc = 0 , sts;

    if (!maxFilecache)
        return 0;

    for (dr = NULL; (dr = getNextRe(objListH, dr));) {
        if ((dr->filter & CAC_PRELOAD)) {
            if ((dir = dirName(dr->raw, 1)) == NULL)
                return -1;
            if (!strlen(dir)) {
                if (defaultDir != NULL) {
                    Free(dir);
                    if ((dir = Strdup(defaultDir)) == NULL)
                        return -1;
                }
            }
            sts = scanDir(dir, &dr->compiled, dr->filter);
            LOG_MSG6 (LOG_DEBUG, "Directory '%s' done", dir);
            Free(dir);
            if (sts < 0)
                return -1;
            rtc += sts;
        }
    }
    return rtc;
}

/*
	load one single file into cache
*/

int preloadFile(void *vdc)
{
    dlc_t *dc = (dlc_t *)vdc;
    ulong oldSize;

    if (cacheLimits((ulong)dc->status.st_size, dc->name))
        return 1;

    oldSize = dc->size;
    if (file2buffer(dc->name, &dc->cache, &dc->size, dc->mark)) {
        LOG_MSG1 (LOG_ERR, _("Cannot load file '%s' into cache: %m"),dc->name);
        deleteCacheEntry(cacheListH, dc);
        return 1;
    }
    LOG_MSG4 (LOG_INFO, _("File '%s' (%lu bytes, '%s') loaded into cache"), 
              dc->name, dc->size, (dc->mark == 2) ? "a" : "b");
    dc->flags |= CAC_VALID;
    dc->used = time(NULL);
    currCachesize -= oldSize;
    currCachesize += dc->size;
    return 0;
}

/*
	load all marked files into cache. candidates are all members of
	chacheListH, with CAC_PRELOAD set and CAC_VALID cleared
*/

int preloadCache(void)
{
    dlc_t *dc;
    int rtc = 0;

    if (!maxFilecache)
        return 0;

    for (dc = NULL; (dc = getNextCache(cacheListH, dc)) != NULL;) {
        if ((dc->flags & CAC_PRELOAD) && !(dc->flags & CAC_VALID))
            rtc += preloadFile(dc);
    }

    rmXferBuffer();
    return rtc;
}


int initCacheBuffer(void *vdc)
{
    dlc_t *dc = (dlc_t *)vdc;

    next  = dc->cache;
    avail = dc->size;
    LOG_MSG6 (LOG_DEBUG, "%s %d bytes in cache", __FUNCTION__, avail);
    return 0;
}


int getCacheBuffer (char *buf, int reqlen)
{
    if (avail < reqlen)
        reqlen = avail;

    memcpy (buf, next, reqlen);
    next  += reqlen;
    avail -= reqlen;
    LOG_MSG6 (LOG_DEBUG, "%s %d bytes from cache, %d left", 
              __FUNCTION__, reqlen, avail);
    return reqlen;
}


int loadCacheFile(char *file, int mode, void **vnew)
{
    dlr_t *obj;
    dlc_t *entry;
    struct stat st;

    obj = findRePattern(objListH, NULL, 0, file, 0);
    if ((obj != NULL) && (obj->filter & CAC_UPDATE)) {
        LOG_MSG5 (LOG_DEBUG, "File is a valid cache object '%s'", obj->raw);
        if (stat(file, &st)) {
            LOG_MSG1 (LOG_ERR, _("Cannot stat file '%s'"), file);
            return 1;
        }
        if (!cacheLimits((ulong)st.st_size, file)) {
            if ((entry = findCacheByMark(cacheListH, file, mode)) != NULL)
                deleteCacheEntry(cacheListH, entry);
            if (addCache(cacheListH,(mode & CAC_MODE), file, &st,
                         (obj->filter & CAC_PROP), &entry)) {
                errNoDynmem;
                return 2;
            }
            if (!preloadFile(entry)) {
                *vnew = entry;
                return 0;
            }

            deleteCacheEntry(cacheListH, entry);
        }
    }
    LOG_MSG5 (LOG_DEBUG, "File not cached");
    return 3;
}
#endif /* FILE_CACHE */
